# glmwidget.py -- widget used to display molecules in 3d
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2006
#                               a-tenderholt@stanford.edu
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

import sys
from math import sqrt,cos,sin,tan,acos,pi

from PyQt4 import QtOpenGL
from PyQt4 import QtGui,QtCore
from OpenGL.GL import *
from OpenGL.GLU import *
from cclib.parser import ccopen
from cclib.bridge import makeopenbabel
import openbabel
import pybel

etable = openbabel.OBElementTable()

LENGTH, ANGLE, TORSION = range(3)

class GLWidget(QtOpenGL.QGLWidget):
    def __init__(self, parent = None):
        QtOpenGL.QGLWidget.__init__(self,parent)
        self.obmol = None
        self.xrot = 0
        self.yrot = 0
        self.zrot = 0
        self.lastPos = QtCore.QPoint()
        self.locked = True
        self.selected = []
        self.selectedLength = []
        self.selectedAngle = []
        self.selectedTorsion = []

    def slotDataChanged(self, topLeft, bottomRight):
        self.updateGL()

    def setSelected(self,list):
        self.selected = list
        self.updateGL()

    def setSelectedType(self, list, type):
        if type == LENGTH:
            self.selectedLength = list
        if type == ANGLE:
            self.selectedAngle = list
        if type == TORSION:
            self.selectedTorsion = list
        
    def slotEnableLock(self,value):
        self.locked = ( value == QtCore.Qt.Checked )

    def mousePressEvent(self, event):
        self.lastPos = QtCore.QPoint(event.pos())

    def mouseMoveEvent(self, event):
        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()
        
        if self.locked:
            self.setXRotation(self.xrot + 8 * dy)
            self.setYRotation(self.yrot + 8 * dx)
        else:
            #determine perp vector
            length = sqrt(dx**2 + dy**2)
            x = -dy/length
            y = dx/length

            angle = length/180 * pi
            costh = cos(angle)
            sinth = sin(angle)

            self.emit(QtCore.SIGNAL("signalRotateMolecule"),angle, [x, y, 0.0])
            
        self.lastPos = QtCore.QPoint(event.pos())
        self.updateGL()

    def orientXYZ(self):
        self.xrot = 0
        self.yrot = 0
        self.zrot = 0

        self.updateGL()

    def orientYXZ(self):
        self.xrot = 0
        self.yrot = 0
        self.zrot = -90 * 16

        self.updateGL()

    def orientXZY(self):
        self.xrot = -90 * 16
        self.yrot = 0
        self.zrot = 0

        self.updateGL()

    def orientZXY(self):
        self.xrot = 90 * 16
        self.yrot = 90 * 16
        self.zrot = 180 * 16

        self.updateGL()

    def setXRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.xrot:
            self.xrot = angle
            self.emit(QtCore.SIGNAL("xRotationChanged(int)"),angle)

    def setYRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.yrot:
            self.yrot = angle
            self.emit(QtCore.SIGNAL("yRotationChanged(int)"),angle)

    def setZRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.zrot:
            self.zrot = angle
            self.emit(QtCore.SIGNAL("zRotationChanged(int)"),angle)

    def normalizeAngle(self, angle):
        while angle < 0:
            angle += 360 * 16
        while angle > 360 * 16:
            angle -= 360 * 16
        return angle

    def setMolecule(self,coords,aonums):
        self.obmol = makeopenbabel(coords,aonums)
        self.updateGL()

    def updateMolecule(self,index,option,value):
        atom = self.obmol.GetAtom(index+1)
        vector = atom.GetVector()

        if option == 0:
            atom.SetAtomicNum(value)

        elif option == 1:
            atom.SetVector(value, vector.y(), vector.z())

        elif option == 2:
            atom.SetVector(vector.x(), value, vector.z())

        elif option == 3:
            atom.SetVector(vector.x(), vector.y(), value)
            
        self.updateGL()
        
    def centerMolecule(self, index):
        atom = self.obmol.GetAtom(index+1)
        vector = atom.GetVector()

        newvector = openbabel.vector3(-1*vector.x(),-1*vector.y(),-1*vector.z())
        self.obmol.Translate(newvector)
        self.updateGL()

    def saveFile(self,filename,format):
        obconv = openbabel.OBConversion()
        ok = obconv.SetOutFormat(format)
        obconv.WriteFile(self.obmol,str(filename))

    def initializeGL(self):

        glEnable(GL_CULL_FACE)
        
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)

        glEnable(GL_NORMALIZE)
        glEnable(GL_BLEND)
        glEnable(GL_LIGHTING)
        glShadeModel(GL_SMOOTH)

        #set ambient and diffuse light colors by glColor
        glEnable(GL_COLOR_MATERIAL)
        glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE) 

        #set intensity of light factors before using glColor
        specularLight = [1.0, 1.0, 1.0, 1.0]
        ambientLight = [0.2, 0.2, 0.2, 1.0]
        diffuseLight = [0.6, 0.6, 0.6, 1.0]
        position = (0.7, 0.8, 0.4, 0.0)

        #set light parameters
        glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight)
        glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight)
        glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight)
        glLightfv(GL_LIGHT0, GL_POSITION, position)
        glEnable(GL_LIGHT0)

        #set relfective properties
        specReflect = [ 0.8, 0.8, 0.8, 1.0]
        glMaterialfv(GL_FRONT, GL_SPECULAR, specReflect)
        glMateriali(GL_FRONT, GL_SHININESS, 45)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glMatrixMode(GL_MODELVIEW)

    def resizeGL(self, width, height):
        side = min(width, height)
        glViewport(0,0,width,height)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60.0, (width+0.0)/height, 0.1, 100.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def paintGL(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        if not self.obmol:
            return

        glLoadIdentity()
        glPushMatrix()
        glTranslatef(0.0, 0.0, -15.0)
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)

        quadric = gluNewQuadric()
        gluQuadricNormals(quadric, GLU_SMOOTH)

        size = self.obmol.NumAtoms()
        mol = pybel.Molecule(self.obmol)
        
        for atom in mol.atoms:
            radius = etable.GetVdwRad(atom.atomicnum)
            rgb = etable.GetRGB(atom.atomicnum)

            glPushMatrix()
            glTranslatef(atom.coords[0], atom.coords[1], atom.coords[2])
            glColor4f(rgb[0], rgb[1], rgb[2], 1.0)
            gluSphere(quadric, radius*0.2, 64, 64)
            glPopMatrix()
        
        for i in range(self.obmol.NumBonds()):

            bond = self.obmol.GetBond(i)

            atoms = [bond.GetBeginAtom(),bond.GetEndAtom()]

            for j in range(2):

                rgb = etable.GetRGB(atoms[j].GetAtomicNum())
                glColor4f(rgb[0], rgb[1], rgb[2],1.0)

                pos1 = pybel.Atom(atoms[j]).coords
                pos2 = pybel.Atom(atoms[1-j]).coords
                
                direction = [ pos2[0] - pos1[0], pos2[1] - pos1[1], pos2[2] - pos1[2] ]

                length = self.length(direction[0], direction[1], direction[2])

                prerot = [ -1*direction[1], direction[0], 0.0 ]
                rotsize = self.length(prerot[0], prerot[1], prerot[2])
                if rotsize != 0:
                    rot = [prerot[0] / rotsize, prerot[1] / rotsize, 0.0] 
                    angle = acos(direction[2]/length)/pi*180
                elif direction[2] < 0:
                    rot = [1.0, 0.0, 0.0]
                    angle = 180
                else:
                    rot = [1.0, 0.0, 0.0]
                    angle = 0

                glPushMatrix()
                glTranslatef(pos1[0], pos1[1], pos1[2])
                glRotatef(angle, rot[0], rot[1], rot[2])
                gluCylinder(quadric, 0.15, 0.15, length/2, 32, 32)
                glPopMatrix()

        glPopMatrix()

#paint axes
        width = 0.06
        length = 0.4
        y = -5.0 * tan(30.0 / 180 * pi)
        glTranslatef(-1.0 * self.width() / self.height() * y, y, -6.0)
        glColor3f(1.0, 1.0, 1.0)
        gluSphere(quadric,width,32,32)

#render x-axis
        glPushMatrix()
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)
        glRotatef(90, 0.0, 1.0, 0.0)

        glColor3f(1.0, 0.0, 0.0)
        gluCylinder(quadric, width, width, length, 32, 32)
        glPushMatrix()
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

#render y-axis
        glPushMatrix()
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)
        glRotatef(-90, 1.0, 0.0, 0.0)

        glColor3f(0.0, 1.0, 0.0)
        gluCylinder(quadric, width, width, length, 32, 32)
        glPushMatrix()
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

#render z-axis
        glPushMatrix()
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)

        glColor3f(0.0, 0.0, 1.0)
        glPushMatrix()
        gluCylinder(quadric, width, width, length, 32, 32)
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

        glLoadIdentity()
        glTranslatef(0.0, 0.0, -15.0)
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)

        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        for i in self.selected:
            atom = pybel.Atom(self.obmol.GetAtom(i+1))
            radius = etable.GetVdwRad(atom.atomicnum)
            
            glPushMatrix()
            glTranslatef(atom.coords[0], atom.coords[1], atom.coords[2])

            glColor4f(1.0, 1.0, 1.0, 0.5)
            gluSphere(quadric,radius*0.3,64,64)

            glPopMatrix()

        for i in self.selectedLength:
            atom = pybel.Atom(self.obmol.GetAtom(i+1))
            radius = etable.GetVdwRad(atom.atomicnum)
            
            glPushMatrix()
            glTranslatef(atom.coords[0], atom.coords[1], atom.coords[2])

            glColor4f(1.0, 0.0, 0.0, 0.5)
            gluSphere(quadric,radius*0.3,64,64)

            glPopMatrix()

        for i in self.selectedAngle:
            atom = pybel.Atom(self.obmol.GetAtom(i+1))
            radius = etable.GetVdwRad(atom.atomicnum)
            
            glPushMatrix()
            glTranslatef(atom.coords[0], atom.coords[1], atom.coords[2])

            glColor4f(0.0, 1.0, 0.0, 0.5)
            gluSphere(quadric,radius*0.3,64,64)

            glPopMatrix()

        for i in self.selectedTorsion:
            atom = pybel.Atom(self.obmol.GetAtom(i+1))
            radius = etable.GetVdwRad(atom.atomicnum)
            
            glPushMatrix()
            glTranslatef(atom.coords[0], atom.coords[1], atom.coords[2])

            glColor4f(0.0, 0.0, 1.0, 0.5)
            gluSphere(quadric,radius*0.3,64,64)

            glPopMatrix()

        gluDeleteQuadric(quadric)

    def length(self, x, y, z):
        return sqrt(x**2 + y**2 + z**2)

if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)
    parser = ccopen(sys.argv[1])
    parser.parse()
    window = GLWidget()
    window.setMolecule(parser.atomcoords[0],parser.atomnos)
    window.resize(800,600)
    window.show()
    sys.exit(app.exec_())

